#include "stdafx.h"
// 																																		TABARI Project
//___________________________________________________________________________________
// 																																	 			input.cp

//  This file contains the routines for handling file input
 
//__________________________________________________________________________________
//
//	 Copyright (c) 2000  Philip A. Schrodt.  All rights reserved.
//
// 	Redistribution and use in source and binary forms, with or without modification,
// 	are permitted under the terms of the GNU General Public License:
// 	http://www.opensource.org/licenses/gpl-license.html
//
//	Report bugs to: schrodt@ku.edu

//	The most recent version of this code is available from the KEDS Web site:
//		http://www.ku.edu/~keds

//	For plausible indenting of this source code, set the tab size in your editor to "2"

//___________________________________________________________________________________
// 																																						Headers

#include "TABARI.h"

//___________________________________________________________________________________
// 																																   external globals

extern TabariFlagsClass TabariFlags;
extern CodeStoreClass CodeStore;
extern CharStoreClass CharStore;
extern RootsClass 		Roots;
extern LiteralsClass 	Literals;  // *** debug only
extern ParserClass		Parser;		
extern CoderClass			Coder;		
extern ProcessorClass  Processor;

//___________________________________________________________________________________
// 																																   internal globals

ReadFilesClass 		ReadFiles;
CommentFileClass	CommentFile;		

//___________________________________________________________________________________
// 																														primary input functions

void ReadFilesClass:: doPhraseError(char *sreason)
// announces error; decides whether to quit; sets the phrase variables to
// defaults if continuing
{
	fPhraseOK = false;
	if (PhraseError(sLine, sreason)) {
//  WriteLine("RFC:dPE: Resetting sPhrase to ***\n",sLine); // ### debug		
		strcpy(sPhrase,sBlankCode);
		strcpy(sCode,sCodeNull);
		sComment[0] = '\0';
	}
	else exit(0);
} // doPhraseError

void ReadFilesClass:: readActors(char *sfilename)
// input routine for .actors file
{
	openFile(sfilename);
	while (readLine()) {
		if (fPhraseOK) {
			if (strstr(sPhrase,"<NOUN")) {
				readWord("</NOUN>",Noun);
				continue;
			}
			if (strstr(sPhrase,"<ADJ")) {
				readWord("</ADJ",Adjctv);
				continue;
			}
			if (sPhrase[0] != ':') {  // [08.06.30] ### this if() is part of the unimplemented agent scheme
				if (Literals.duplicateRoot(sPhrase)) {
					strcat(sPhrase,"\" <ACTOR> has already been entered in the dictionary");
					ShowWarningError("\"",sPhrase,"New entry has been skipped",shError00);
				}
				else {
					try {Roots.storeActor();}
					catch (int i) { doPhraseError((char*)sErrEMPTY_CODE);}
				}
			}
//		else Roots.store(InPhrase.&sPhrase[1],Agent);// ### more unimplemented agent code
		}
	}
	storeTail(Actor);

} // readActors

void ReadFilesClass:: readVerbs(char *sfilename)
// input routine for .verbs file
{
toktype idxroot;
instring sverb; // current verb being processed
bool gotVerb = false; // checks against possibility of verbless pattern (it's happened...)
	
	openFile(sfilename);
	while (readLine()) {
		try {
			if (fPhraseOK) {
	//			cout << sPhrase << endl;
				if (strstr(sPhrase,"<TIME>")) {
					readTime();
					continue;
				}
				if (strstr(sPhrase,"<ATTRIB")) {
					readAttrib();
					continue;
				}
				if (sPhrase[0] != '-') {
					if (Literals.duplicateRoot(sPhrase)) {
						strcpy(sverb,sPhrase);
						strcat(sPhrase,"\" <VERB> has already been entered in the dictionary");
						ShowWarningError("\"",sPhrase,"Verb has been skipped; any included patterns will cause an error",shError00);
						gotVerb = false;
						strcpy(sPhrase, "Duplicate: ");  // signal that verb was a duplicate
						strcat(sPhrase,sverb);
						strcpy(sverb,sPhrase);	
					}
					else {	idxroot = Roots.storeVerb();
						if (fPhraseOK) {   // ### can storeVerb detect an error?
	//					  if (sPhrase[0] == '*') {
	//						    WriteLine("RFC:rV: resetting sverb to ***\n",sLine); // ### debug
	//								WriteNewLine(sPhrase);
	//						}
							strcpy(sverb,sPhrase);
							gotVerb = true;
						}
					}
				}
				else {
					if (gotVerb && strchr(sPhrase,'*')) Roots.storePattern(idxroot,Verb);
					else {
						if (gotVerb) throw MISS_VERBM;
						else {
							if (sverb[9] == ':') throw DUP_VERB;  //  a generally safe if not flawless test for the duplicate condition
							else throw PAT_ORPHAN;
						}
					}
				}
			}
		}
		catch (int i) {
			ResetCurs();
			WRITESTRING("\nPattern error in .verbs file for verb \"");
			WRITESTRING(sverb);
			WRITESTRING("\"");
			WRITEEOL();
			switch (i) {
				case MISS_LBRAC: doPhraseError((char*)sErrMISS_LBRAC); break;
				case MISS_RBRAC: doPhraseError((char*)sErrMISS_RBRAC); break;
				case MISS_VERBM: doPhraseError("No verb marker \'*\' has been specified"); break;				
				case EMPTY_CODE: doPhraseError((char*)sErrEMPTY_CODE); break;				
				case PAT_ORPHAN: ShowWarningError("Patterns found at beginning of .verb file","A verb must come before any patterns",shError00); break;
				case DUP_VERB: ShowWarningError("Verb for this pattern duplicated an existing root","Remove one of the duplicate entries",shError00); break;
			}
		}
	}
	storeTail(Verb);
} // readVerbs

void ReadFilesClass:: readTime(void)
// input routine for <TIME>...</TIME> segment of verbs file
{
toktype idxroot;
	
	TabariFlags.fhasTimeList = true;
	
	while (readLine()) {
		if (fPhraseOK) {
//			cout << sPhrase << endl;
			if (strstr(sPhrase,"</TIME>")) return;  // normal exit point
			if (sPhrase[0] != '-') idxroot = Roots.storeTime();
			else Roots.storePattern(idxroot,Time);
		}
	}
	ShowFatalError("No </TIME> tag in file ",curFilename,shError00);
} // readTime

void ReadFilesClass:: readAttrib(void)
// input routine for <ATTRIB>...</ATTRIB> segment of verbs file
{
toktype idxroot;
	
	TabariFlags.fhasAttribList = true;
	fcodeOptional = true;  // this prevents parseline from indicating error due to no code
	
	while (readLine()) {
		if (fPhraseOK) {
//			cout << sPhrase << endl;
			if (strstr(sPhrase,"</ATTRIB")) {
				fcodeOptional = false;
				return;  // normal exit point
			}
			if (sPhrase[0] != '-') idxroot = Roots.storeAttrib();
			else Roots.storePattern(idxroot,Attrib);
		}
	}
	ShowFatalError("No </ATTRIB... tag in file ",curFilename,shError00);
} // readTime

void ReadFilesClass:: readWord(char *sendtag, wordtype wtype)
// input routine for <NOUN> and <ADJECTIVE> segments of actors file
{	
	fcodeOptional = true;  // this prevents parseline from indicating error due to no code
	if (Noun == wtype) TabariFlags.fhasNounList	= true;
	else if (Adjctv == wtype) TabariFlags.fhasAdjctvList	= true;

	while (readLine()) {
		if (fPhraseOK) {
//			cout << sPhrase << endl;
			if (strstr(sPhrase,sendtag)) {
				fcodeOptional = false;
				return;  // normal exit point for while(readLine()) loop
			}
			if (Literals.duplicateRoot(sPhrase)) {
				strcat(sPhrase,"\" <");
				strcat(sPhrase,&sendtag[2]);
				strcat(sPhrase,"> has already been entered in the dictionary");
				ShowWarningError("\"",sPhrase,"New entry has been skipped",shError00);
			}
			else Roots.storeWord(sPhrase,wtype);
		}
	}
	ShowFatalError("No </NOUN> or </ADJ... tag in file ",curFilename,shError00); // make general
} // readTime

//___________________________________________________________________________________
// 																																	          issues

int ReadFilesClass:: getISSUE(void)
// process the ISSUES tag; returns current value of kategory
// This thing should eventually be back-fitted with a real XML parser; at the moment we
// are just faking it...  ###
// The CATEGORY text is saved but at the moment we don't do anything
// isNumeric is set to true if issue is numeric
{
static issueHeadStruct * pHead = &Coder.issueHead; // current issue header
static int curkat = 0;  			// current category
int 	 kat = curkat++;
char * pstext = sComment;		// use the class input globals for temporary storage
char * pst;

																// set up issue header
	if (pHead->kategory >= 0) {		// create a new link at end (note that pHead is static)
		pHead->pnext = new issueHeadStruct;
		pHead = pHead->pnext;
		if (!pHead) ShowFatalError("Unable to allocate memory for new issue in ReadFilesClass:: readIssues",sPhrase,shError11);
		pHead->found = false;
		pHead->pnext = NULL;
		pHead->plist = NULL;
	}
	pHead->kategory = kat;
	fisNumeric = false;

	getField(sPhrase,"CATEGORY",pstext); 	// run this first to make sure the syntax is okay
	if (*pstext) {												// but actually do the transfer with an erase
		pst = strstr(sPhrase,"CATEGORY"); 
		pst += 8;
		while (*pst != '\"') ++pst; // move to the "
		++pst;			// move past the "
		while (*pst != '\"') {  // copy text inside quotes into sComment
			*pstext++ = *pst; 
			*pst++ = '-';     // erase the text so we don't confuse it with a keyword
		}
	  *pstext = '\0';
	  pHead->phrase = CharStore.putCharStr(sComment);
  }
  else pHead->phrase = CharStore.charNullString;

	pstext = sComment;  // continue to use this for intermediate storage
	getField(sPhrase,"TYPE",pstext); // check for "TYPE=" field
	if (*pstext) {
		pHead->doAll    = (bool) strstr(pstext,"ALL");
		pHead->doNumber = (bool) strstr(pstext,"NUM");
		fisNumeric = pHead->doNumber;
	}
	else {
		pHead->doAll		= false;
		pHead->doNumber = false;
	}

	getField(sPhrase,"DEFAULT",pstext); // check for "DEFAULT=" field
	if (*pstext) pHead->icode = 	CodeStore.addCode(pstext);
	else pHead->icode = 0;
	
	return kat;					
} // getISSUE

void ReadFilesClass:: readIssues(char *sfilename)
// input routine for .issues file
// Note that lines between <ISSUE>...</ISSUE> will be skipped; this is controlled by kat < 0
{
int 		kat = -1;

	fisNumeric = false;   // also set in the constructor...
	openFile(sfilename);
	while (readLine()) {
		if (fPhraseOK) {
			if (strstr(sPhrase,"<ISSUE")) kat = getISSUE();
			else if (strstr(sPhrase,"</ISSUE")) {
				 kat = -1;
				 fisNumeric = false;
			}
			else if (kat >= 0) {
				try {Roots.storeIssue(kat,fisNumeric);}
				catch (int i) {
					if (MISS_LBRAC == i) doPhraseError((char*)sErrMISS_LBRAC);	
					else doPhraseError((char*)sErrMISS_RBRAC);
				}	
			}
		}
	} // while
} // readIssues

//___________________________________________________________________________________
// 																																	            FREQs

int ReadFilesClass:: getFREQ(void)
// process the FREQ tag; returns current value of kategory
{
static freqStruct * pHead = &Coder.freqHead; // current issue header
int 	 kat = Coder.iFreq++; // current category -- starts at 1
char * pstext = sComment;		// use the class input globals for temporary storage
char * pst;

																// set up issue header
	if (pHead->kategory >= 0) {		// create a new link at the end of the list (note that pHead is static)
		pHead->pnext = new freqStruct;
		pHead = pHead->pnext;
		if (!pHead) ShowFatalError("Unable to allocate memory for new issue in ReadFilesClass:: readFreq",sPhrase,shError11);
		pHead->pnext = NULL;
	}
	pHead->kategory = kat;

	getField(sPhrase,"CATEGORY",pstext); 	// run this first to make sure the syntax is okay
	if (*pstext) {												// but actually do the transfer with an erase
		pst = strstr(sPhrase,"CATEGORY"); 
		pst += 8;  // skip past known text
		while (*pst != '\"') ++pst; // move to the "
		++pst;			// move past the "
		while (*pst != '\"') {  // copy text inside quotes into sComment
			*pstext++ = *pst; 
			*pst++ = '-';     // erase the text so we don't confuse it with a keyword
		}
	  *pstext = '\0';
	  pHead->phrase = CharStore.putCharStr(sComment);
  }
  else pHead->phrase = CharStore.charNullString;

	pstext = sComment;
	getField(sPhrase,"ABBREV",pstext); 	// run this first to make sure the syntax is okay
	if (*pstext) {												// but actually do the transfer with an erase
		pst = strstr(sPhrase,"ABBREV"); 
		pst += 6;
		while (*pst != '\"') ++pst; // move to the "
		++pst;			// move past the "
		while (*pst != '\"') {  // copy text inside quotes into sComment
			*pstext++ = *pst; 
			*pst++ = '-';     // erase the text so we don't confuse it with a keyword
		}
	  *pstext = '\0';
//	  cout << "ABBREV: " << sComment << endl; // ***
	  pHead->abbrev = CharStore.putCharStr(sComment);
  }
  else pHead->abbrev = CharStore.charNullString;
	
	return kat;					
} // getFREQ

void ReadFilesClass:: readFreqs(char *sfilename)
// input routine for .freq file
// Note that lines between </FREQ>...<FREQ> will be skipped; this is controlled by kat < 0
{
int 		kat = -1;

	fdoingFreq	= true;
	openFile(sfilename);
	while (readLine()) {
		if (fPhraseOK) {
			if (strstr(sPhrase,"<FREQ")) kat = getFREQ();
			else if (strstr(sPhrase,"</FREQ")) {
				 kat = -1;
			}
			else if (kat >= 0) {
//				cout << "rf: " << sPhrase << "k = " << kat << endl;
				Literals.storeFreq(sPhrase,kat);
			}
		}
	} // while
	fdoingFreq	= false;
} // readFreqs

//___________________________________________________________________________________
// 																																	working functions

void ReadFilesClass:: doLINE_LONG(void)  
// handles LINE_LONG exception
{
instring serr;
	strcpy(serr,"Input line too long in file ");
	strcat(serr, curFilename);
	strcat(serr,"\nProblem line:\n");
	ShowWarningError(serr,sLine,"Line will be ignored",shError32);
	sLine[0] = '/';  // force line to be skipped
}	

void ReadFilesClass:: insertStr(char *s1, char *s2)  
// inserts s2 at the beginning of s1, overwriting the first char of s1
// This is typically called with s1 set in the middle of the string
// Note: this is used to replace single delimiters in the input with 
//       space-separated delimiters
{
	char *s = s2;
	char *sa = s1;
	int len;
//	char *sc = s1; // *** debugging: used to track original string
	
	while (*s) s++;  		// find end of s1, s2
	while (*sa) sa++; 
	len = (int)(s - s2) - 1;		// number of cells we need to open up
	s = sa;				// move string back len cells
	sa += len;
	while (s > s1)  *sa-- =  *s--;  // copy all but first char
	while (*s2) *s1++ = *s2++; 	// copy s2 into the open cells
} // insertStr

void ReadFilesClass:: getsLine(void)
// safe version of getline: throws LINE_LONG exception if line is too long.
// Finishes reading line before throwing exception 
{
char * ps = sLine;
int ka = 0;

	while (ka++ < MAX_TAB_INPUT) {
		fin.get(*ps);
		if (('\n' == *ps) || (fin.eof())) {
			*ps = '\0';
			return;
		}
		++ps;
	}
	--ps;
	while (('\n' != *ps) && (!fin.eof())) {  // finish up reading line
		fin.get(*ps);
	}
	*ps = '\0';  // terminate string so it can be displayed
	throw LINE_LONG;

} // getsLine

void ReadFilesClass:: openFile(char *sfilename)
{	
	fin.open(sfilename,ios::in); 
	if (fin.fail()) ShowFatalError("Unable to open the input file ",sfilename,shError03);
	WriteLine("Initializing from ",sfilename);
	strcpy(curFilename,sfilename);
} // open

bool ReadFilesClass::readLine(void)
// Generic phrase reading procedure that handles all of the file input;
// stores comment info if appropriate }
{	
	do {
		try {
			getsLine();
//		cout << sLine << endl; // DEBUG ***
		}
		catch (int i) { doLINE_LONG(); }	
	}
	while ((!fin.eof()) && ((!sLine[0]) || ( '/' == sLine[0]) || ( '#' == sLine[0])));	

	if ((fin.eof()) || ('~' == sLine[0])) {
		fPhraseOK = false;
		return false;
	}

	TrimBlanks(sLine);
	parseLine();	// pull out the various parts of the input

	if (sComment[0]) locComment = CommentFile.put(sComment); 	// handle comments 
	else locComment = -1;  																		// no comments on this line
	
	return true;
} // readLine

void ReadFilesClass:: storeTail(wordtype wtype)
// store the tail lines in CommentFile, then close fin
{
	while (!fin.eof()) {
		try { 
			getsLine(); 
		}
		catch (int i) {  // handles LINE_LONG exception
			instring serr;
			strcpy(serr,"Final comments line too long in file ");
			strcat(serr, curFilename);
			strcat(serr,"\nProblem line:\n");
			ShowWarningError(serr,sLine,"Line will be truncated",shError32);
		}
		if (fin.eof()) break;
		if (fin.fail()) ShowFatalError("Input problem in ReadFilesClass::storeTail",shError05);
		CommentFile.putTail(sLine,wtype);
	}
	fin.clear(); 	// clear eof flag 
	fin.close(); 
	if (fin.fail()) ShowFatalError("Unable to close the input file",shError04);
} // storeTail

void ReadFilesClass::addBlanks (char *s, char c, int &len)
// adds blanks around c; increments len; does nothing if len>=MAX_TAB_INPUT
// This is currently only used for {, |, and }, and therefore has idiosyncratic
// treatment of each
{
	char *pst = s;
	char si[4] = "   ";
	bool done;
	
	if (len >= MAX_TAB_INPUT) return;
	while ((pst) && (pst = strchr(pst,c))) {		// the assignment is deliberate here
		if ((*(pst-1) != ' ') || (*(pst+1) != ' ')) {
			done = false;
			if (('{' == c) && ('_' == *(pst-1))) {  // allow _{
				strcpy(si,"{ ");
				insertStr(pst,si);
				done = true;
			}
			else if (('}' == c) && ('_' == *(pst+1))) { // allow }_
				strcpy(si," }");
				insertStr(pst,si);
				done = true;
			}
			if (done) strcpy(si,"   ");
			else {
				si[1] = c;
				insertStr(pst,si);
			}
			pst += 2;
			len += 2;
		}
		else if (pst) ++pst;
	}
} // addBlanks

void ReadFilesClass::parseLine (void)
// Generic phrase parsing procedure that handles all phrase [code] strings
// Input is through sLine 
// Sets the values of sCode, sPhrase,and sComment
{
instring sa;
char 		*pst;
int 		len = 0;  // used to check total length of sPhrase

	fPhraseOK = false;

	if (!sLine[0]) {
		doPhraseError("Blank line");
		return;
	}
	if (strchr(sLine,';')) Copy2Chr(sa,sLine,';');			// get non-comment part of s 
	else strcpy(sa,sLine);
	
	if (fdoingFreq || ('<' == sa[0])) {   // XML tag or freqs; just return in sPhrase
		strcpy(sPhrase,sa);
		sComment[0] = '\0'; //  no comments on this line 
		sCode[0] = '\0';		// no code either
		fPhraseOK = true;		
		return;
	}		

	pst = strchr(sa,'[');
	if (pst) {
		bool gotContent = false;
		Copy2Chr(sPhrase,sa,'[');
		strcpy(sCode,pst);
		pst = sCode + 1 ;
		while (!gotContent && (*pst != ']') && *pst) {  // check for non-blank content
			if (*pst > 32) gotContent = true;
			++pst;
		}
		if (!gotContent) {
			doPhraseError((char*)sErrEMPTY_CODE);
			return;
		}
	}
	else {
		strcpy(sPhrase,sa);
		strcpy(sCode,sNullCode);
		if ((!fcodeOptional) && (!fisNumeric)) {
			doPhraseError((char*)sErrMISS_LBRAC);  // allow empty code on numeric issues
			return;
		}
	}

	pst = sPhrase;			// shift to upper case
	while (*pst) {
		*pst = toupper(*pst); 
		++pst;
		++len;
	}
// delimit '{', '}' and '|'
	addBlanks(sPhrase,'{',len);  
	addBlanks(sPhrase,'|',len);
	addBlanks(sPhrase,'}',len);
	if (len >= MAX_TAB_INPUT)  doPhraseError("Phrase too long after adding blanks to {, |, }");
	
	TrimEndBlanks(sPhrase);
  if ((!sPhrase[0]) || (!sPhrase[1])) doPhraseError("No phrase has been specified");  // ### i.e. phrase must be at least two chars in length -- check that this is the correct contingency, e.g. A_ is as short as we go

	pst = strchr(sLine,';');
	if (pst) strcpy(sComment,++pst);	// get the comment from end of line
	else sComment[0] = '\0';  // else no comments on this line 

	fPhraseOK = true;

} // parseLine 


//___________________________________________________________________________________
// 																																		.options file

void ReadFilesClass:: optionsError (const char *s)
{
	WriteLine("This .options command has a problem: ",sLine);
	ShowWarningError(s,"Line was skipped",shError42);
} // optionsError

void ReadFilesClass:: trimCommand (char *s)
// trim command from s and delete leading blanks
{	
char *ps = strchr(s,':') + 1;   // ':' exists, as this was checked earlier

	while (isspace(*ps)) ++ps;  // skip over blanks as well
	while (*ps) *s++ = *ps++;		// delete first part of string
	*s = '\0';									// terminate string	

} // trimCommand

void ReadFilesClass:: getField (char *s, char *substr, char *sout)
// Gets the ="..." in s following substr, assigns blank-trimmed value to sout. 
// If substr is not present or there is no number, sets iout = 0 and 
// sends ShowWarning about the syntax error
{
char * ps = strstr(s,substr);		// find start of substring ;
char * psa;

	if (!ps) {			// substring wasn't found
		*sout = '\0';
		return;
	}
	
	try {
		while ((*ps) && ('=' != *ps)) ++ps; // go to the next =
		if (!*ps) throw 1;

		while ((*ps) && ('\"' != *ps)) ++ps; // go to the next "
		if (!*ps) throw 2;

		++ps;
		psa = ps;   		// make sure we've got a closing " before copying
		while ((*psa) && ('\"' != *psa)) ++psa; // go to the next "
		if (!*psa) throw 3;

		psa = sout;  			// save the start of sout
		while ((*ps) && ('\"' != *ps)) *sout++ = *ps++; // it's okay, so copy into sout
		*sout = '\0';			// terminate the string
		TrimBlanks(psa);	// blank-trim sout
		return;
	} // try
	
	catch (int i) {  // syntax error
		char serror[32];
		WriteLine("XML field in this line has a problem: ",sLine);
		switch (i) {
		 	case 1: 	strcpy(serror,"No = following ");
		 						break;
		 	case 2: 	strcpy(serror,"No opening \" following ");
		 						break;
		 	case 3: 	strcpy(serror,"No closing \" following ");
		 						break;
		}
		ShowWarningError(serror,substr,"Field will be ignored.",shError42);
		*sout = '\0';
		return;
	} // catch
		
} // getInteger

void ReadFilesClass:: getInteger (char *s, char *substr, int &iout)
// Gets the first number in s following substr, assigns value to iout. 
// If substr is not present or there is no number, sets iout = 0
{
char * ps = strstr(s,substr);		// find start of substring ;
char * psa = substr;

	iout = 0;									// set iout to default
	if (!ps) return;					// substring isn't there, so quit					
	
	while (*(++psa)) ++ps; // move past the end of the substring
	++ps;
		
	while ((*ps) && !isdigit(*ps)) ++ps;  // go to first digit
	if (!*ps) return;											// no number, so default
	iout = atoi(ps);											// get the integer

} // getField

bool ReadFilesClass:: getBoolean(void)
// looks for a T/F after an =; defaults to true.  Assumes the input
//	has been shifted to upper case 
{
char * ps = strchr(sLine,'=');

	if (ps) {
		while (*ps) {
			if ('T' == *ps) return true;
			else if ('F' == *ps) return false;
			else ++ps;
		}
	}
	else optionsError("Missing '=' in command");
	return true;
} // getBoolean

int ReadFilesClass:: get3Way(void)
// looks for a OFF/T/F after an =; defaults to OFF.  Assumes the input
//	has been shifted to upper case 
{
char * ps = strchr(sLine,'=');
	if (ps) {   // ### why am I rewriting strchr()? <07.08.03>
		if (strstr(ps,"OFF")) return OFF_3WAY;
		while (*ps) {
			if ('T' == *ps) return TRUE_3WAY;
			else if ('F' == *ps) return FALSE_3WAY;
			else ++ps;
		}
	}
	else optionsError("Missing '=' in command");
	return OFF_3WAY;
} // get3Way


void ReadFilesClass:: getCOMPLEX(void)
// process the COMPLEX command 
{
char *ps;

	trimCommand(sLine);
	ps = sLine-1;
	while (*(++ps)) *ps = toupper(*ps);	// convert entire line to upper case

	if ('V' == sLine[0]) {  // deals with an initial 'VERBS'
		sLine[0] = 'X';
		getInteger(sLine, "XERBS", Parser.cplxLimit.nVerb);
	}
	else  getInteger(sLine, " VERBS", Parser.cplxLimit.nVerb); // leading space is used to avoid hitting misspelled NOVERBS, LATEVERBS

	getInteger(sLine, "ACTORS", Parser.cplxLimit.nActor);
	getInteger(sLine, "PRONOUNS", Parser.cplxLimit.nPronoun);
	getInteger(sLine, "CONJ", Parser.cplxLimit.nConj);
	getInteger(sLine, "COMMA", Parser.cplxLimit.nNonRes);
	getInteger(sLine, "LATEVERB", Parser.cplxLimit.nLateVerb);
	Parser.cplxLimit.fPreAct = strstr(sLine,"ACTPRIOR");
	Parser.cplxLimit.fPostAct = strstr(sLine,"ACTAFTER");
	Parser.cplxLimit.fNoSrc = strstr(sLine,"NOSOURCE");
	Parser.cplxLimit.fNoVerb = strstr(sLine,"NOVERB");
	Parser.cplxLimit.fNoTar = strstr(sLine,"NOTARGET");
	Parser.cplxLimit.fNoEvt = strstr(sLine,"NOEVENT");
	Parser.fCplxStrg = strstr(sLine,"EXPLAIN");

int ka = Parser.cplxLimit.nVerb + Parser.cplxLimit.nActor + Parser.cplxLimit.nPronoun + 
				 Parser.cplxLimit.nConj + Parser.cplxLimit.nNonRes + Parser.cplxLimit.nLateVerb;

bool fok = Parser.cplxLimit.fPreAct || Parser.cplxLimit.fPostAct || Parser.cplxLimit.fNoSrc || 
					 Parser.cplxLimit.fNoVerb || Parser.cplxLimit.fNoTar || Parser.cplxLimit.fNoEvt;

	Parser.fCheckCplx = (ka > 0) || fok;
} // getCOMPLEX

void ReadFilesClass:: getEQUIV(void)
// process the EQUIV command 
{
char *ps;

	trimCommand(sLine);
	ps = sLine-1;
	while (*(++ps)) *ps = toupper(*ps);	// convert entire line to upper case
	
// currently doesn't do anything
} // getEQUIV

void ReadFilesClass:: getFORWARD(void)
// process the format string in the FORWARD command
{	

char * ps;
char * pt;

	trimCommand(sLine);
	ps = sLine;
	if (!(*sLine)) return;
	
	if (strstr(sLine," DOC")) fhasDoc = true;
	else fhasDoc = false;
	
	while ((*ps) && ('^' == *ps)) ++ps;  // skip initial blanks
	if (*ps) {  // set IDInfo location 
	//		if ka > 1 then
	//			Delete(strin, 1, ka - 1);
		iIDInfoLoc = (int)(ps - sLine);
		pt = ps;
		while ((*pt) && (' ' != *pt)) ++pt;  // terminate at blank (end of record ID, before DOC or OFF)
		*pt = '\0';
		iIDInfoLen = strlen(ps);
		if (iIDInfoLen > MAX_RECID_LEN) { 		// trim to maximum length 
				iIDInfoLen = MAX_RECID_LEN;
				ShowWarningError("FORWARD string  was >31 chars: ",ps,"String was truncated",shError41);
				*(ps+MAX_RECID_LEN) = '\0';
		}
	}			
	else {
		ShowWarningError("FORWARD string had no non-blank characters","FORWARD option was not processed",shError42);
		return ;
	}

	ps = strchr(sLine,'N');
	if (ps) { 												// set story serial # location 
		iStoryLoc = (int)(ps - sLine);
		pt = ps + 1;
		while ((*pt) && ('N' == *pt)) ++pt;
		iStoryLen = (int) (pt - ps);
	}

	ps = strchr(sLine,'S');
	if (ps) { 											// set sequence serial # location 
		if (!strstr(sLine,"OFF")) Parser.fuseForward = true;
		iSentLoc = (int)(ps - sLine);
		pt = ps + 1;
		while ((*pt) && ('S' == *pt)) ++pt;
		iSentLen = (int) (pt - ps);
	}

} // getFORWARD

void ReadFilesClass:: getOUTPUT(void)
// process the OUTPUT command 
{
	fOutputID = (bool)(strstr(sLine, "ID"));
	fOutputStory = (bool)(strstr(sLine, "STO"));;
	fOutputSeq = (bool)(strstr(sLine, "SEN"));;
	fOutputDoc = (bool)(strstr(sLine, "DOC"));;
	fOutputText = (bool)(strstr(sLine, "TEX"));;
	fOutputPats = (bool)(strstr(sLine, "PAT"));;
} // getOUTPUT

void ReadFilesClass:: getCOMMA(void)
// process the COMMA command 
{
} // getCOMMA

void ReadFilesClass:: getLABEL(void)
// process the LABEL command 
{
char * ps = strchr(sLine,':') + 1; // readOptions has checked that the ':' is present
char * pscode = sCode;				// use the class input globals here
char * pstext = sPhrase;

	while ((*ps) && (isspace(*ps))) ++ps; // skip over any leading white space
	if (!*ps) {
		optionsError("No code specified in LABEL command");
		return;
	}

	while ((*ps) && (*ps != '=') && (!isspace(*ps))) *pscode++ = *ps++; // copy the code
	if (*ps) *pscode = '\0';
	else {
		optionsError("No '=' in LABEL command");
		return;
	}

	while ((*ps) && (('=' == *ps) || (isspace(*ps)))) ps++; // skip the '=' and any white space
	if (!*ps) {
		optionsError("No label following '=' in LABEL command");
		return;
	}

	while (*ps) *pstext++ = *ps++; // copy the rest of the line into the label
	*pstext = '\0';
	TrimEndBlanks(sPhrase);

	CodeStore.addCode(sCode,sPhrase);  // add the code and label
} // getLABEL

void ReadFilesClass:: getSET(void)
// process the SET command 
{
char *ps;

	trimCommand(sLine);
	ps = sLine-1;
	while (*(++ps)) *ps = toupper(*ps);	// convert entire line to upper case
	
	if      (strstr(sLine, "CODE"))  Coder.setcodeMode(sLine);
	else if (strstr(sLine, "HAIKU")) TabariFlags.fShowHaiku = getBoolean();
	else if (strstr(sLine, "YYYY")) TabariFlags.f4DigitYear = getBoolean();
	else if (strstr(sLine, "LABEL")) TabariFlags.fAddLabels = getBoolean();
	else if (strstr(sLine, "MATCH")) TabariFlags.fAddMatch  = getBoolean();
	else if (strstr(sLine, "FBIS"))  TabariFlags.fFBISMode  = getBoolean();
	else if (strstr(sLine, "MINIM")) {
		getInteger (sLine, "MINIM",Parser.minLex); // MINIMUM WORDS
		--Parser.minLex;   // reduce by one so we can compare to iLex
	}
	else if (strstr(sLine, "USAGE")) {
		if (strstr(sLine, "ACT"))      TabariFlags.fActorUsage = get3Way();
		else if (strstr(sLine, "VER")) TabariFlags.fVerbUsage = get3Way();
		else optionsError("Unrecognized SET: USAGE command");
	}
	else optionsError("Unrecognized SET: command");

} // getSET

void ReadFilesClass:: readOptions(char *sfilename)
// input routine for .options file
{
char *ps;

	openFile(sfilename);
	while (!fin.eof()) {
		do {
			try {
				getsLine();
			}
			catch (int i) { doLINE_LONG(); }
		}
		while ((!fin.eof()) && ((!sLine[0]) || ( '/' == sLine[0]) || ( '#' == sLine[0])));		// '//' and # is the comment indicator

		if ((fin.eof()) || ('~' == sLine[0])) {
			fin.close();
			return;
		}
		
		ps = strchr(sLine,':');
		if (!ps) {
			optionsError("No ':' in command");
			continue;
		}
		
		ps = sLine - 1;
		while (*(++ps) != ':') 	*ps = toupper(*ps); // convert command to upper case

		switch (sLine[0]) {

			case 'C': switch (sLine[3]) {
									case 'M': getCOMMA();
														break;
									case 'P': getCOMPLEX();
														break;
									default : optionsError("Unrecognized command");
								}
								break;

			case 'E': getEQUIV();
					  		break;

			case 'F': getFORWARD();
					  		break;

			case 'L': getLABEL();
					  		break;

			case 'O': getOUTPUT();
					  		break;

			case 'S': getSET();
								break;

			case 'V': if ('A' == sLine[1]) { // VALID
									while (*(++ps)) *ps = toupper(*ps); 			// convert remainder of line upper case
									Coder.fSrcValid = strstr(sLine,"SOURCE");
									Coder.fTarValid = strstr(sLine,"TARGET");
									Coder.fShowInvalid = strstr(sLine,"DISPLAY");
								}
								break;
			
			default: optionsError("unrecognized command");
		}
	} // while
	fin.close();
	return;

} // readOptions

//___________________________________________________________________________________
// 																														class CommentFileClass

int CommentFileClass::intType(wordtype wtype)
// returns an index number for wtype
{
	switch (wtype) {
	case Actor : return 0;
	case Verb  : return 1;
	case Synym : return 2;
	default    : return -1; // this shouldn't occur
	} // wtype
} // intType

commloc CommentFileClass::put(char *sphrase)
 // stores comment; returns pointer
{
commloc loc = fcomm.tellp();

	while (*sphrase) {				// put sphrase into fcomm
		fcomm.put(*sphrase);
		if (fcomm.fail()) ShowFatalError("Output problem in CommentFileClass::put",shError06);
		++sphrase;
		}
		fcomm.put('\n');  // terminate with a newline so we can read this
	return loc;
	
} // put
 
void CommentFileClass::get(commloc loc, char *scomm)
// retrieves comment, puts in sphrase
{
	fcomm.seekg(loc);
	fcomm.getline(scomm, MAX_TAB_INPUT);
	if (fcomm.fail()) ShowFatalError("Input problem in CommentFileClass::get",shError05);
} // get

void CommentFileClass::putTail(char *sphrase, wordtype wtype)
// stores tail information -- this is the information following the ~~ FINISH
// terminator in the .actors or .verbs file.  Then closes fin
{
int itype = intType(wtype);
		
	if (tailLoc[itype] < 0) tailLoc[itype] = fcomm.tellp();  // record the start of the tail
	put(sphrase);
	++tailLen[itype];		
} // putTail

void CommentFileClass::reopen(wordtype wtype)
// close fcomm; re-opens as an ios:in file; sets theTail to appropriate file type
{
	int it = intType(wtype);
	
	curLen = tailLen[it];
	curLoc = tailLoc[it];
	
	fcomm.close(); 
	if (fcomm.fail()) ShowFatalError("Can't close comment file in CommentFileClass::reopen",shError04);
	fcomm.open(sfcommname,ios::in); 
	if (fcomm.fail()) ShowFatalError("Unable to reopen comments file",shError07);
} // reopen

bool CommentFileClass:: getTail(char *s)
// returns ~~ FINISH line and tail information.  "reopen" needs to be called first 
// note the somewhat tricky use of curLoc to determine if this is the first call
{		
	if (curLoc >= 0) {		// first time this has been called; return the ~~FINISH line
		litstring sdate,stime;
		MakeTimeDateStr (stime, sdate);			// get time and date strings
		strcpy(s,"~~ FINISH ");
		strcat(s,stime);
		strcat(s,sdate);
		fcomm.seekg(curLoc);  										// go to the start of the tail
		curLoc = -1;															// signal that we've done this
		return true;
	}
	else if (--curLen > 0) {		// read the tail, decrementing curLen
		fcomm.getline(s, MAX_TAB_INPUT);
		if (fcomm.fail()) ShowFatalError("Input problem in CommentFileClass::getTail",shError05);
		return true;
	}
	else return false;		// got everything

} // getTail	

//___________________________________________________________________________________
// 																																		debugging code

#if FALSE
void ReadFilesClass:: testOptions(void)
// *** debugging : tests assorted .options functions
{
instring stest;
	
	strcpy(stest, "COMPLEX: COMPLEX[2] DISCARD[2]");
	trimCommand(stest);
		
// getInteger ("COMPLEX: COMPLEX[2] DISCARD[2]","DISCARD", irec);
// getInteger ("COMPLEX: COMPLEX[2] DISCARD [213]","DISCARD", irec);

} // testOptions
#endif

#if FALSE
void ReadFilesClass:: test(void) 
// *** debugging function 
{
	strcpy(sLine," * {OUT|OFF|DOWN}[012]");
//	strcpy(sLine,"* OF + IN $  [212] ;JON 5/26/95");
	cout << "Input  :" << sLine << endl;
	parseLine();
	cout << "Phrase :" << sPhrase << endl;
	cout << "Code   :" << sCode << endl;
	cout << "Comment:" << sComment << endl;
	cout << "OK?    :" << fPhraseOK << endl;

} // test

void ReadFilesClass:: set(char *sphrase, char *scode, char *scomment, commloc loc)   
// sets the local variables
// *** currently just used for debugging
{
	strcpy(sPhrase, sphrase);
	strcpy(sCode, scode);
	strcpy(sComment, scomment);
	locComment = loc;
	fPhraseOK = true;
} // set

#endif


