#include "stdafx.h"
#include <conio.h> //getch() not found otherwise ?
// 																																		TABARI Project
//__________________________________________________________________________________
// 																																			interface.cp

// 	This file contains an assortment of general-purpose utilities that involve the
// 	screen interface.  This contains all of the screen input except for the functions 
//  found in modify.cp.  Unlike the GUI	interface of KEDS, TABARI  uses a very simple 
//  keyboard-driven "dumb terminal" interface.  This is designed to work on a screen 
//	that is 48 lines high by 128 characters wide.
//
//	Screen output is  generally are isolated in a limited number of functions in 
//	the various classes -- these usually have a "show" prefix and only use cout  
//	or addstr() for output.
//
//  Note that most of the *file* I/O is in the class ProcessorClass.
//
//	Operating System Options
//	These are set using the NCURSES, LINUX, and WINDOWS globals in TABARI.h.  
//	One and only one can be true; otherwise the program will terminate in the 
//	constructor for TabariFlagsClass.
//
//	Input is isolated in the functions getUpper, ReadLine and GetInput: these are
//	configured so that the program responds immediately to keystrokes in menus
//	("noncanonical", in Unix terminology) but waits for a return (and recognizes
//	the backscape/delete key) when entering codes and phrases.  How this (and the
//	echoing of keystrokes) is handled depends on the operating system.
//
//	Output is done through a number of basic output macros and functions, plus
//  some interface-specific code for more elaborate structures such as menus
//  and list displays.
//
//	#define NCURSES = TRUE
//	For versions 0.5 and later, this is the standard interface. It implements a 
//  terminal interface using the gcc ncurses package; it should run without 
//	changes in the OS-X "Terminal", Linux, and X-Windows. As of version 0.6,
//  this is the only version of the program that is being tested consistently.
//
//	#define LINUX = TRUE
//  This was the Linux version for versions earlier than 0.5 and was a bit of
//  a hack since the program was actually developed in using  the CodeWarrior 
//  SIOUX input/output window.  It is currently still in the code in order
//  to provide <iostream> output in addition to ncurses output for debugging
//  purposes, but is only tested occasionally. ncurses works just fine in 
//  Linux, there isn't any good reason to use this.
//
//	The screen is initially set to noncanonical mode in the initTerminal function; 
//	the characteristics of the original screen (which is presumably canonical) are 
//	saved.  Whenever a string needs to be entered, the program toggles out of
//	canonical mode, gets the string, then toggles back into canonical mode.
//
//	From the user's perspective, this should be seamless.  The only potentially
//	problematic side-effect would be if the program crashed and left the terminal
//	in canonical mode -- a reboot may be required in this instance.
//
//	#define WINDOWS = TRUE
//	Never thought you'd see this out of the KEDS project, eh?  Actually, at the
//	moment it hasn't been implemented, but it in here as a stub and exercise for 
//  someone else. As of August 2007, still no volunteers
//
//  Note: versions prior to 0.6 contained code in a #define MAC = TRUE section
// 	that used the CodeWarrior SIOUX input/output window from the <sioux.h> library, 
//	which provides a scrolling window and cut, paste and save from a menu. This
//  has been removed in version 0.6 since that library is no longer supported.


//__________________________________________________________________________________
//
//	Copyright (c) 2002-2007  Philip A. Schrodt.  All rights reserved.
//
// 	Redistribution and use in source and binary forms, with or without modification,
// 	are permitted under the terms of the GNU General Public License:
// 	http://www.opensource.org/licenses/gpl-license.html
//
//	Report bugs to: schrodt@ku.edu

//	The most recent version of this code is available from the KEDS Web site:
//		http://web.ku.edu/keds

//	For plausible indenting of this source code, set the tab size in your editor to "2"

//___________________________________________________________________________________
// 																																			Headers

#include "TABARI.h"

#if LINUX
#include <stdio.h>
#include <unistd.h>
//#include <termios.h> // already in TABARI.h
#define SYSVERSION "<iostream> version for Unix and Linux\n"
#endif

#if NCURSES 
#define SYSVERSION "Unix \"ncurses\" version for Macintosh OS-X and Linux\n"
#else
 #define SYSVERSION "Windows Visual Studio C++ Express 2008\n"
#endif


//___________________________________________________________________________________
// 																								Function prototypes not in Tabari.h

void InitTABARI(void);				// Do introduction
void initTerminal(void);			// Set Linux terminal parameters
void firstHaiku(void);				// selects and displays introductory haiku
void writeIntro(void);				// Writes the introductory screen

extern TabariFlagsClass 	TabariFlags;

//___________________________________________________________________________________
// 																																			

#if LINUX 

void initTerminal(void)
// initializes screen to a non-canonical terminal so that menus don't require a <enter>
// follows the example in the Linux info:libc#Noncanon_Example
// ### old Linux version
{

  if (!isatty(STDIN_FILENO)) {
  	ShowFatalError("Screen is not a Linux 'terminal' device",shError00); 
  	exit(0);
  }
 	tcgetattr(STDIN_FILENO,&TabariFlags.savedSettings);  // save terminal characteristics
 		
  tcgetattr(STDIN_FILENO,&TabariFlags.keyedSettings);
  TabariFlags.keyedSettings.c_lflag &= ~(ICANON);  // clear ICANON and ECHO
 TabariFlags.keyedSettings.c_cc[VMIN] = 1;
  TabariFlags.keyedSettings.c_cc[VTIME] = 0;
  tcsetattr(STDIN_FILENO,TCSAFLUSH,&TabariFlags.keyedSettings);
  
} // initTerminal

char getch(void)
// return a char from the keyboard
// This has been replaced with the ncurses equivalent 
{
char c;
 	fflush(stdout);
	read(STDIN_FILENO, &c,1);
	return c;
}

char getUpper(void)
// returns uppercase keystroke without echo; converts \n to \r
{
char	c = toupper(getch());
	cout << endl;
	if ('\n' == c) c = '\r';  // convert Linux \n to a Mac \r
	return c;
}

void ClearScreen(void)
// Linux mode uses a scrolling screen, so this just puts in a divider.
{ 
int ka;
	cout << '\n';
	for (ka = 0; ka < WINDOW_WIDTH; ++ka) cout << '=';
	cout << "\n\n";
}
 
void InitTABARI(void)		
// Do introduction
{
	initTerminal();
	writeIntro();
}

void ReadLine(char *s)
// reads s from keyboard.  In Linux, it toggles first to the canonical
//  status so that the editing keys work, reads the string, then back to non-canonical
{ 
 	fflush(stdout);
   	tcsetattr(STDIN_FILENO,TCSANOW,&TabariFlags.savedSettings); 	
	cin.getline(s,MAX_TAB_INPUT); 
	tcsetattr(STDIN_FILENO,TCSANOW,&TabariFlags.keyedSettings);
	
} // ReadLine/Linux

void GetInput(char *s, char *sprompt)
// write sprompt on new line, then read s from keyboard In Linux, it toggles first to the canonical
//  status so that the editing keys work, reads the string, then back to non-canonical
{ 
	tcsetattr(STDIN_FILENO,TCSANOW,&TabariFlags.savedSettings);
	cout << '\n' << sprompt << ": "; 
 	fflush(stdout);
	cin.getline(s,MAX_TAB_INPUT);
	tcsetattr(STDIN_FILENO,TCSANOW,&TabariFlags.keyedSettings);

} //GetInput/Linux

#endif   // LINUX

#if NCURSES

void initTerminal(void)
// initializes the ncurses environment
{
	initscr(); 
	cbreak(); 
	noecho();
	nonl();
	intrflush(stdscr, FALSE);
	keypad(stdscr, TRUE);
} // initTerminal

char getUpper(void)
// returns uppercase keystroke without echo; converts \n to \r
{
char	c = toupper(getch());
	if ('\n' == c) c = '\r';  // convert Linux \n to a Mac \r
	return c;
}

void ClearScreen(void)
// clears screen but does not refresh
{ 
	erase();
	TabariFlags.cursc = 0;
	TabariFlags.cursr = 0;
 }
 
void InitTABARI(void)		
// Do introduction
{
	initTerminal();
	writeIntro();
}

void ResetCurs(void)
// refresh screen, then get cursr, cursc. Used following series of screen
// writes where final location of cursor is indeterminant
{
	refresh();
	getyx(stdscr,TabariFlags.cursr,TabariFlags.cursc);
}

void ReadLine(char *s)
// reads s from keyboard.
// ncurses: toggles to the canonical status, reads the string,
// then back to non-canonical
{ 
	nocbreak();
	echo();
	getstr(s);
	refresh(); // is this needed? ###
	getyx(stdscr,TabariFlags.cursr,TabariFlags.cursc);
	cbreak(); 
	noecho();
	} // ReadLine/ncurses

void GetInput(char *s, char *sprompt)
// write sprompt on current line, then read s from keyboard 
// ncurses: toggles to the canonical status, reads the string,
// then back to non-canonical
{ 
	nocbreak();
	echo();
	addstr(sprompt);
	addstr(": ");
	refresh();
	getstr(s);
	getyx(stdscr,TabariFlags.cursr,TabariFlags.cursc);
	cbreak(); 
	noecho();
} //GetInput/ncurses

bool safeGetLine(char *s)
// gets a line of text from keyboard, checking that buffer isn't exceeded.
// Note:This function gets around an infinite loop that occurred when a user typed
//      a string that exceeded the length of instring s.  I would think that there
//      would be a straightforward way around this using standard stdio functions, 
//      but I couldn't figure it out, and it might be necessary because I'm using 
//      <return>-free keyboard input in the MetroWerks SIOUX environment.  If you 
//      know a more elegant solution, please implement it... [02.02.07]

{
char *ps = s;
int   ka = 0;

	while (ka < MAX_TAB_INPUT) {
		*ps = getch();
		addch(*ps);
		if (('\r' == *ps) || ('\n' == *ps)) { // handle Mac and Linux <returns>
			break;
		}
		++ka;
		++ps;
	} 
	if (ka >= MAX_TAB_INPUT) {
		WriteLine("Input line length exceeded; please hit <return> at the end of each line");
		s[MAX_TAB_INPUT-1] = '\0';
		return false;
	}
	*ps = '\0';
	return true;
} // safeGetLine

// NOTE: All of the Write*() functions require that TabariFlags.cursr is set
//       correctly; if mixing these with the WRITE*() macros, use ResetCurs() 
//       first      

void WriteNewLine(char *s)
// write s to display on new line
{
	getyx(stdscr,TabariFlags.cursr,TabariFlags.cursc);
	move(++TabariFlags.cursr,0);
	addstr(s);
	move(++TabariFlags.cursr,0);
	refresh();
}

void WriteLine(char *s)
// write s to display
{
//	addstr("###");  ###
	addstr(s);
	move(++TabariFlags.cursr,0);
	refresh();
}

void WriteLine(const char *s)
// write s to display
{
//	addstr("---");  ###
	addstr(s);
	move(++TabariFlags.cursr,0);
	refresh();
}

void WriteLine(char *s1, char *s2)
// write s1 and s2 to display
{
//	addstr("***");  ###
	addstr(s1);
	addstr(s2);
	move(++TabariFlags.cursr,0);
	refresh();
}

void WriteLine(char *s1, const char *s2)
// write s1 and s2 to display
{
//	addstr("+++");  ###
	addstr(s1);
	addstr(s2);
	move(++TabariFlags.cursr,0);
	refresh();
}

void WriteAlert(char *s1)	
// beep, then write s1  to display
{
	beep();
	getyx(stdscr,TabariFlags.cursr,TabariFlags.cursc);
	move(++TabariFlags.cursr,0);
	addstr(s1);
	move(++TabariFlags.cursr,0);	
	refresh();
}

void WriteAlert(const char *s1)	
// beep, then write s1  to display
{
	beep();
	getyx(stdscr,TabariFlags.cursr,TabariFlags.cursc);
	move(++TabariFlags.cursr,0);
	addstr(s1);
	move(++TabariFlags.cursr,0);	
	refresh();
}

void WriteAlert(char *s1, char *s2)	
// beep, then write s1 and s2 to display
{
	beep();
	getyx(stdscr,TabariFlags.cursr,TabariFlags.cursc);
	move(++TabariFlags.cursr,0);
	addstr(s1);
	addstr(s2);
	move(++TabariFlags.cursr,0);	
	refresh();
}

void WriteLong(char *s1, long i)
// write s1, integer and \n without refresh
{
char sout[16];
	addstr(s1);
	sprintf(sout,"%ld",i);
	addstr(sout);
	move(++TabariFlags.cursr,0);
}

void WriteFloat(char *s1, float r)
// write s1, float and \n without refresh
{
char sout[16];
	addstr(s1);
	sprintf(sout,"%10.2f",r);
	addstr(sout);
	move(++TabariFlags.cursr,0);
}

bool GetAnswer(char *sprompt, char cyes, char cno)
// writes sprompt, reads a character, returns true if this is upper or lower cyes;
// response must be cyes or cno
{
char c;

	while (true) {
		beep();
		addstr(sprompt);
		addstr(" ->");
		refresh();
		c = toupper(getch());
		addch(c);
		refresh();
		if ((toupper(cno) == c) || (toupper(cyes) == c)) break;
		beep();
  	mvaddstr(++TabariFlags.cursr,0,"Please enter ");
  	addch(cyes);
  	addstr(" or ");
  	addch(cno);
  	move(--TabariFlags.cursr,0);	
  	clrtoeol();	
		--TabariFlags.cursr;
	}	
	WRITEEOL();
	if (c == toupper(cyes)) return true;
	else return false;
} // GetAnswer

bool GetAnswer(char *s1, char * s2)
// two strings, "Y" is the affirmative response; response must be 'N' or 'Y'
{
char c;

	while (true) {
		beep();
		addstr(s1);
		addstr(s2);
		addstr("? (Y/N) ->");
		refresh();
		c = toupper(getch());
		addch(c);
		refresh();
		if (('N' == c) || ('Y' == c)) break;
		beep();
  	mvaddstr(++TabariFlags.cursr,0,"Please enter Y or N");
  	move(--TabariFlags.cursr,0);	
  	clrtoeol();	
		--TabariFlags.cursr;
	}	
	WRITEEOL();
	if (c == 'Y') return true;
	else return false;
} // GetAnswer

void Pause(void)
// wait for char; exit if Q
{
	WriteNewLine("Pause: press any key to continue; Q to exit program without saving changes");
	char c = getUpper();
	if ('Q' == c) {
		addstr("Exiting...");
		refresh();
		exit(0);
	}
} // Pause


void showHaiku(const char * shaiku)
// Display a haiku string
{
const char * pst = shaiku;

	if (!*pst) return;
	addstr("\n--  ");
	while ((*pst) && ('/' != *pst)) addch(*pst++);
	if (!*pst) {
		addch('\n');
		showHaiku(shError99);
		return;
	}
	addstr("\n--   ");
	pst++;
	while ((*pst) && ('/' != *pst)) addch(*pst++);;
	if (!*pst) {
		addch('\n');
		showHaiku(shError99);
		return;
	}
	addstr("\n--    ");
	addstr(++pst);
	addch('\n');
	ResetCurs();
		 
} // showHaiku

void showMenu(char *s)
// display menu prompt
{
		mvaddstr(++TabariFlags.cursr,0,s);
		refresh();
} // showMenu

void correctMenu(char *s)
// prompts with correct response, returns to previous line
{
	beep();
	mvaddstr(++TabariFlags.cursr,0,s);
	move(--TabariFlags.cursr,0);
	clrtoeol();
	--TabariFlags.cursr;
} // correctMenu

void setupMenu(bool hasinit)
// move menu to a specific location on screen; call with false to set location
{
static int promptrow;

	if (hasinit) {
		move(promptrow+1,0);
		clrtoeol();
		TabariFlags.cursr = promptrow; // note that showMenu will increment this
//		mvaddstr(TabariFlags.cursr,0,">>>>>>>>>>>");
	}
	else {
		ResetCurs();
		promptrow = TabariFlags.cursr; 
		hline(0,WINDOW_WIDTH); // use default horz line char
//		mvaddstr(TabariFlags.cursr,0,"<<<<<<<<<<<<");
	}
} // setupMenu

int AutocodeSetup(void)	
// setup autocoding options
{
char c;
	ClearScreen();
	if (TabariFlags.fAutoCodemode > -1) return TabariFlags.fAutoCodemode; // level was set in command line
	while (true) {
		beep();
		addstr("-- Automatic coding mode --\n");
		addstr("\nFeedback level: N)one  D)ate  F)ull  H)elp  C)ancel ->");
		switch (getUpper()) {
			case '\r':;						// handles <enter>
			case 'N': return 0;
			case 'D': return 1;
			case 'F': return 2;
			case 'H': ShowAutoHelp();
			case 'X':;
			case 'Q':;
			case 'C': return 3;
		}
		if (c != 'H') {
			beep();
			addstr("\nPlease enter N, D, F, H or C\n");
		}
	}
// ### if a complex file is being written, confirm that user still wants it here	
} // autoSetup

#else // ---------------------------------------------------------------------------
// Older versions using <iostream>

void InitTABARI(void)		
// Do introduction
{
	//initTerminal(); //Windows
	writeIntro();
}

void ClearScreen(void)
// clears screen but does not refresh
{ 
	//erase();
	//TabariFlags.cursc = 0;
	//TabariFlags.cursr = 0;
	return;
 }

char getUpper(void)
// returns uppercase keystroke without echo; converts \n to \r
{
//char	c = toupper(getch());
//	if ('\n' == c) c = '\r';  // convert Linux \n to a Mac \r
	return toupper(getch()); //c;
}

void GetInput(char *s, char *sprompt)
{// write sprompt on current line, then read s from keyboard 
 cout << sprompt;
 cin>>s;

}

void ResetCurs(void)
// does nothing in <iostream> mode
{
	return;
}

bool safeGetLine(char *s)
// gets a line of text from keyboard, checking that buffer isn't exceeded.
// Note:This function gets around an infinite loop that occurred when a user typed
//      a string that exceeded the length of instring s.  I would think that there
//      would be a straightforward way around this using standard stdio functions, 
//      but I couldn't figure it out, and it might be necessary because I'm using 
//      <return>-free keyboard input in the MetroWerks SIOUX environment.  If you 
//      know a more elegant solution, please implement it... [02.02.07]

{
char *ps = s;
int   ka = 0;

	while (ka < MAX_TAB_INPUT) {
		*ps = getch();
		cout << *ps;
		if (('\r' == *ps) || ('\n' == *ps)) { // handle Mac and Linux <returns>
			cout << endl;
			break;
		}
		++ka;
		++ps;
	} 
	if (ka >= MAX_TAB_INPUT) {
		cout << "Input line length exceeded; please hit <return> at the end of each line\n";
		s[MAX_TAB_INPUT-1] = '\0';
		return false;
	}
	*ps = '\0';
	return true;
} // safeGetLine


void WriteNewLine(char *s1)
// write s1 to display on new line
{
	cout << '\n' << s1 <<endl; 
}

void WriteLine(char *s1)
// write s1 to display
{
	cout << s1 <<endl; 
}

void WriteLine(const char *s1)
// write s1 and s2 to display
{
	cout << s1 <<endl; 
}

void WriteLine(char *s1, char *s2)
// write s1 and s2 to display
{
	cout << s1 << s2 <<endl; 
}

void WriteLine(char *s1, const char *s2)
// write s1 and s2 to display
{
	cout << s1 << s2 <<endl; 
}

void WriteAlert(char *s1)	
// beep, then write s1  to display
{
	cout << '\a' << s1 << endl;
}

void WriteAlert(const char *s1)	
// beep, then write s1  to display
{
	cout << '\a' << s1 << endl;
}

void WriteAlert(char *s1, char *s2)	
// beep, then write s1 and s2 to display
{
	cout << '\a' << s1 << s2 << endl;
}

void WriteLong(char *s1, long i)
// write s1, integer and \n without refresh
{
	cout << s1 << i << endl;
}

void WriteFloat(char *s1, float r)
// write s1, float and \n without refresh
{
	cout << s1 << r << endl;
}

bool GetAnswer(char *sprompt, char cyes, char cno)
// writes sprompt, reads a character, returns true if this is upper or lower cyes;
// response must be cyes or cno
{
char c;

	while (true) {
		cout << '\a' << sprompt << " ->";
		c = toupper(getch());
		cout << c << endl;
		if ((toupper(cno) == c) || (toupper(cyes) == c)) break;
		cout << "\aPlease enter " << cyes << " or " << cno << endl;
	}	
	if (c == toupper(cyes)) return true;
	else return false;
} // GetAnswer

bool GetAnswer(char *s1, char * s2)
// two strings, "Y" is the affirmative response; response must be 'N' or 'Y'
{
char c;

	while (true) {
		cout << '\a' << s1 << s2 << "? (Y/N) ->";
		c = toupper(getch());
		cout << c << endl;
		if (('N' == c) || ('Y' == c)) break;
		cout << "\aPlease enter Y or N" << endl;
	}	
	if (c == 'Y') return true;
	else return false;
} // GetAnswer

void Pause(void)
// wait for char; exit if Q
{
	cout << "Pause: press any key to continue; Q to exit program without saving changes" << endl;
	char c = getUpper();
	if ('Q' == c) {
		cout << "Exiting...";
		exit(0);
	}
} // Pause

void showHaiku(const char * shaiku)
// Display a haiku string
{
const char * pst = shaiku;

	if (!*pst) return;
	putchar('\n');
	putchar('-'); putchar('-'); putchar(' '); putchar(' '); 	
	while ((*pst) && ('/' != *pst)) putchar(*pst++);
	if (!*pst) {
		putchar('\n');
		showHaiku(shError99);
		return;
	}
	putchar('\n'); 
	putchar('-'); putchar('-'); putchar(' '); putchar(' ');
	putchar(' ');
	pst++;
	while ((*pst) && ('/' != *pst)) putchar(*pst++);
	if (!*pst) {
		putchar('\n');
		showHaiku(shError99);
		return;
	}
	putchar('\n'); 
	putchar('-'); putchar('-'); putchar(' '); putchar(' ');
	putchar(' ');	putchar(' ');
	puts(++pst);
	putchar('\n');
		 
} // showHaiku

void setupMenu(bool hasinit)
// move menu to a specific location on screen; does nothing in <iostream> mode
{
	return;
} // setupMenu

void showMenu(char *s)
// display menu prompt
{
		cout << s;
} // showMenu

void correctMenu(char *s)
// prompts with correct response, returns to previous line
{
	cout << "\n\a" << s << endl;
} // correctMenu

int AutocodeSetup(void)	
// setup autocoding options
{
char c;
	ClearScreen();
	while (true) {
		cout << "\a-- Automatic coding mode --\n";
		cout << "\nFeedback level: N)one  D)ate  F)ull  H)elp  C)ancel ->";
		switch (getUpper()) {
			case '\r':;						// handles <enter>
			case 'N': return 0;
			case 'D': return 1;
			case 'F': return 2;
			case 'H': ShowAutoHelp();
			case 'X':;
			case 'Q':;
			case 'C': return 3;
		}
		if (c != 'H') cout << "\n\aPlease enter N, D, F, H or C\n";
	}
// ### if a complex file is being written, confirm that user still wants it here	
} // autoSetup

#endif

bool PhraseError (char *stin, char *sreason)
// Announces error in .verbs and .actors input, clear screen after response
// Called from: parsePhrase
// Returns false if quitting
{
int ka;
int startrow = TabariFlags.cursr;  // save where we've started

	WriteAlert("Error in the input line");
	WriteLine(stin);
	WriteLine("Problem: ",sreason);
	ka = GetAnswer("Enter C to continue, Q to quit",'C','Q');
#if NCURSES   // ### need to work around this at some point...
	move(startrow,0);
	clrtobot(); // clear the message
#endif
	TabariFlags.cursr = startrow;
	return ka;	
} // phraseError

void ShowWarningError (const char *sreason, const char *sresponse, const char *shaiku)
// Announces an error, waits for <return>, then erases itself and returns.
// Entering q+<return> will exit the program
{
int startrow = TabariFlags.cursr; // save where we've started
  
	WriteAlert("Warning! Program is encountering a potential problem");
	WriteLine("Problem : ",sreason);
	WriteLine("Response: ", sresponse);
	showHaiku(shaiku); 
	WriteLine("Press <Return> to continue processing or Q to quit.");
	if ('Q' == toupper(getch())) exit(1);  // do-what-I-mean programming...
  else {
#if NCURSES   // ####
		move(startrow,0);
		clrtobot(); // clear the message
#endif
	}
} // ShowWarningError

void ShowWarningError (const char *sreason, char *s2, const char *sresponse, const char *shaiku)
// Version of function that concatenates two strings
// (this is typically used to add a phrase or file name to the message) 
// Entering q+<return> will exit the program
{
int startrow = TabariFlags.cursr; // save where we've started

	WriteAlert("Warning! Program is encountering a potential problem");
	WRITESTRING("Problem : ");
	WRITESTRING(sreason);
	WriteLine(s2);
	WriteLine("Response: ", sresponse);
	showHaiku(shaiku); 
	WriteLine("Press <Return> to continue processing or Q to quit.");
	if ('Q' == toupper(getch())) exit(1);  
  else {
#if NCURSES // ####
		move(startrow,0);
		clrtobot(); // clear the message
#endif
	}
} // ShowWarningError

void ShowFatalError (const char *sreason, const char *shaiku)
// Announces an error, waits for <return>, then calls exit(1)
{
	WriteAlert("Fatal error encountered in program");
	WriteLine("Problem : ",sreason);
	showHaiku(shaiku); 
	WriteLine("Press <Return> to end processing.");
	getch();
	exit(1);		// ### we don't actually *use* the '1' at the moment...
} // ShowFatalError

void ShowFatalError (const char *sreason, char *sfile, const char *shaiku)
// Same as above but with two strings; usually the second is a file name
{
	WriteAlert("Fatal error encountered in program");
	WriteLine("Problem : ",sreason);
	WRITESTRING(sreason);
	WriteLine(sfile);
	showHaiku(shaiku); 
	WriteLine("Press <Return> to end processing.");
	getch();
	exit(1);		// ### we don't actually *use* the '1' at the moment...
} // ShowFatalError

int MenuMain(void)
// Options at program start
{
	while (true) {
		showMenu("Select: C)oding  D)ifferences  H)elp  Q)uit ->");
		switch (getUpper()) {
		case 'Q': return 0;
		case 'D': return 1;
		case 'H': return 2;
		case '\r':;						// handles <enter>
		case 'C': return 3;
		}
		correctMenu("Please enter C, D, H, or Q");
	}	
} // MenuMain

int RecordMenu(void)
// Menu for the primary read-parse-code loop
{
	while (true) {
		setupMenu(true);
		showMenu("Select: N)ext  P)arsing  M)odify  R)ecode  A)utocode  O)ther  Q)uit ->");
		switch (getUpper()) {
		case '\r':;						// handles <enter>
		case ' ' :;						// because my thumb rests on this key...
		case 'N': return 0;		// what the documentation actually says
		case 'P': return 1;
		case 'M': return 2;
		case 'R': return 3;
		case 'A': return 4;
		case '0': ;						// zero
		case 'O': return 5;
		case 'X': ;
		case 'Q': return 6;
		case '<': return 7;  // reverse
		case ',': return 7;
		}
		correctMenu("Please enter N, P, M, R, A, or Q");
	}	
} // RecordMenu

int OtherMenu(bool hasProb)
// Menu for the O)ther option in doProcess
{
	while (true) {
		if (hasProb) showMenu("Select: B)ackup  P)roblem  M)emory  U)sage  eX)it ->");
		else  showMenu("Select: B)ackup  M)emory  U)seage  eX)it ->");
		switch (getUpper()) {
		case '\r':;						// handles <enter>
		case 'E': ;
		case 'X': ;
		case 'Q':
#if NCURSES
							move(TabariFlags.cursr,0);  // erase the menu
							clrtoeol();
#endif
							return 0;
		case 'B': return 1;
		case 'P': if (hasProb) return 2;
							else break;
		case 'M': return 3;
		case 'U': return 4;
		case 'V': return 5; // doValidation: invisible to the user
		case 'T': return 6; // showTags: invisible to the user
		}
		if (hasProb) correctMenu("Please enter B, P, M, U, or X");
		else correctMenu("Please enter B, M, U, or X"); 
	}	
} // OtherMenu

bool ReadArg(int argc, char *argv[],  char *sfilename)		// Process command line
	// Process command line arguments
{
	if (argv[1][0] != '-') {
		strcpy(sfilename, argv[1]);
		return 1;
	}
	else {
	if (argv[1][1] != 'a') {
		ShowWarningError("Unrecognized argument in command line","Command line input will be ignored",shError00);
		return 0;
	}
		switch (argv[1][2]) {  // currently only handles -ax option   
			case 'n' :
			case '\0':  TabariFlags.fAutoCodemode = 0; break;  // default and 'None'  ### replace the magic numbers here; also let 0 indicate OFF...
			case 'd' :  TabariFlags.fAutoCodemode = 1; break;  // 'Date'
			case 'f' :  TabariFlags.fAutoCodemode = 2; break;  // 'Full'
			default  :  ShowWarningError("Unrecognized argument in command line","Command line input will be ignored",shError00);
									return 0;
		} // ### error check other options
		strcpy(sfilename, argv[2]);
		return 1;
	}
		
} // ReadArg

void firstHaiku(void)
// selects and displays introductory haiku
{
	char str[8];
	time_t cur_time;
	int ka;
	int nseas;
	
	time(&cur_time);
	ka = cur_time % (MAX_SEASON + MAX_INTRO);
	if (ka < MAX_INTRO) showHaiku(shIntro[ka]);
	else {
		strncpy(str,ctime(&cur_time),8);
		switch (str[4]) {
			case 'J': if ('u' == str[5]) nseas = 2; else nseas = 0;
				break;
			case 'A': if ('p' == str[5]) nseas = 1; else nseas = 2;
				break;
			case 'F':
			case 'D': nseas = 0;
				break;
			case 'M': nseas = 1;
				break;
			default: nseas = 3; // S, O, N
		}
		switch (nseas) {
			case 0:  showHaiku(shWinter[ka - MAX_INTRO]);
				break;
			case 1:  showHaiku(shSpring[ka - MAX_INTRO]);
				break;
			case 2:  showHaiku(shSummer[ka - MAX_INTRO]);
				break;
			case 3:  showHaiku(shAutumn[ka - MAX_INTRO]);
		}
	}
} // firstHaiku

void writeIntro(void)
// Writes the introductory screen
{
	WRITESTRING("\n\n");
	WRITESTRING("TTTTTTTT     A      BBBBB       A      RRRRR    II\n");
	WRITESTRING("   TT       A A     B    B     A A     R    R   II\n");
	WRITESTRING("   TT      A   A    B    B    A   A    R    R   II\n");
	WRITESTRING("   TT     AAAAAAA   BBBBB    AAAAAAA   RRRRR    II\n");
	WRITESTRING("   TT     A     A   B    B   A     A   R  R     II\n");
	WRITESTRING("   TT     A     A   B    B   A     A   R   R    II\n");
	WRITESTRING("   TT     A     A   BBBBB    A     A   R    R   II\n\n");
	WRITESTRING("Text Analysis By Augmented Replacement Instructions\n");
	WRITESTRING("Version ");
	WRITESTRING(sRelease);
	WRITESTRING(" (");
	WRITESTRING(sReleaseDate);
	WRITESTRING(")\n\n");
	WRITESTRING("Kansas Event Data System Project\n");
	WRITESTRING("Center for International Political Analysis\n");
	WRITESTRING("University of Kansas\nLawrence, Kansas 66045, U.S.A.\n");
	WRITESTRING("http://www.ku.edu/~keds\n\n");
	WRITESTRING("Redistribution and use of TABARI in source and binary forms, with or without \n");
	WRITESTRING("modification, are permitted under the terms of the GNU General Public License:\n");
	WRITESTRING("http://www.opensource.org/licenses/gpl-license.html\n\n");
  WRITESTRING(SYSVERSION); // macro is defined at beginning of interface.cp
	WRITESTRING("Please set your terminal window to display 120 characters by 48 lines\n");
	firstHaiku();
	ResetCurs();
} // writeIntro

void HelpMain(void)
// A primitive help screen -- this is going to get generalized...
{
	WRITESTRING("\nHelp\n\n");
	WRITESTRING("TABARI: Version 0.05 Initial Menu Options:\n\n");
	WRITESTRING("C)ode will request the name of a TABARI project file and then code the text\n");
	WRITESTRING("  recordsfor that project.  Enter <return> without a file name to run the \n");
	WRITESTRING("  demonstration file.\n");
	WRITESTRING("D)ifference runs the dictionary comparison routine\n");
	WRITESTRING("H)elp gives this screen\n");
	WRITESTRING("Q)uit ends the program\n\n");
	WRITESTRING("At the menu prompt, type the appropriate key -- C, D, H or Q -- to select \n");
	WRITESTRING("an option.  <Return> selects the first option in the menu\n\n");
	ResetCurs();
	Pause();
}

void ShowAutoHelp(void)	
// Autocoding help screen
{
	ClearScreen();
	WRITESTRING("Autocoding Help\n\n");
	WRITESTRING("Autocoding codes the entire sets of files specified in the project without\n");
	WRITESTRING("pausing.  Once the process has been initiated, it cannot be stopped except\n");
	WRITESTRING("by terminating the program using Cmd-. (period) or Cmd-Q.\n\n");
	WRITESTRING("The speed of autocoding is dramatically affected by the amount of screen \n");
	WRITESTRING("feedback -- the less that is written to the screen, the faster the coding.\n");
	WRITESTRING("There are three levels\n\n");
	WRITESTRING("N)one.  No feedback on individual records; text file names are displayed as\n");
	WRITESTRING("        they are processed. [default] \n\n");
	WRITESTRING("D)ate.  The date and record ID of each record is displayed as the record is\n");
	WRITESTRING("        processed.  This is useful in identifying a record that is causing\n");
	WRITESTRING("        the system to crash. \n\n");
	WRITESTRING("F)ull.  Display the text and the coded events as in standard coding.\n\n");
	WRITESTRING("H)elp.  Shows this screen\n\n");
	WRITESTRING("C)ancel.Exit the autocoding mode.\n\n");
	WRITESTRING("At the menu prompt, type the appropriate key -- N, D, F, H or C -- to select \n");
	WRITESTRING("an option.  <Return> selects the default option, N)one.\n\n");
	Pause();
	ClearScreen();
} // ShowAutoHelp

