#include "stdafx.h"
// 																																		TABARI Project
//___________________________________________________________________________________
// 																																	 	mergefiles.cp

//  This file contains the routines for comparing a new dictionary with an existing
//	dictionary.
//
//	NOTE: This was written fairly quickly and is not particularly elegant -- several
//        things are done brute-force (for example, the entire pattern list is 
//        checked rather than terminating when the pattern is too long)
 
//__________________________________________________________________________________

//	 Copyright (c) 2000 - 2005 Philip A. Schrodt.  All rights reserved.
//
// 	Redistribution and use in source and binary forms, with or without modification,
// 	are permitted under the terms of the GNU General Public License:
// 	http://www.opensource.org/licenses/gpl-license.html
//
//	Report bugs to: schrodt@ku.edu

//	The most recent version of this code is available from the KEDS Web site:
//		http://www.ku.edu/~keds

//	For plausible indenting of this source code, set the tab size in your editor to "2"

//___________________________________________________________________________________
// 																																						Headers

#include "TABARI.h"

//___________________________________________________________________________________
// 																																   external globals

extern CodeStoreClass 	CodeStore;
extern RootsClass 			Roots;
extern ReadFilesClass 	ReadFiles;
extern LiteralsClass 		Literals;
extern PhrasesClass 		Phrases;

//___________________________________________________________________________________
// 																																   internal globals

MergeFilesClass 		MergeFiles;

//___________________________________________________________________________________
// 																																  

void MergeFilesClass:: doMerge(void)
// driver routine for dictionary comparison
{
bool fActor;
bool fMore = true;
instring sfout;
	
	ClearScreen();
	WriteLine("Welcome to the TABARI dictionary difference routine");
	WriteLine("To use the difference function, enter the name of the reference file (the file");
	WriteLine("you want to compare against), and an output file.  You can then loop through");
	WriteLine("multiple comparison files.  The output file will contain a listing of any");
	WriteLine("new roots, phrases or code changes in the comparison file.\n");
	WriteLine("  If you are comparing dictionaries produced for different coding systems, ");
	WriteLine("answer N to the query \"Check for differences in codes?\".\n");
	fActor = GetAnswer("Enter A to compare .actors files, V to compare .verbs files",'A','V');
	GetInput(srefilename,"\nEnter name of reference file");
	if (fActor) ReadFiles.readActors(srefilename);
	else				ReadFiles.readVerbs(srefilename);

	while (fMore) {					// open output file
		GetInput(sfout,"Enter name of output file");
		fout.open(sfout,ios::in); 		// check if this exists
		if (!fout.fail()) {
			fMore = !GetAnswer("Okay to write over ",sfout);
			if (fMore) continue;
		}
		fout.close();
		fout.clear();
		fMore = false;
		fout.open(sfout,ios::out); 
		if (fout.fail()) {		
			ShowWarningError("Disk probably full or locked -- can't open file ",sfout, "Comparison procedure cancelled",shError00);
			return;
		}
	}
	
	fMore = true;
	while (fMore) {
		fout << "\nReference file  : " << srefilename << endl;
		WriteLine("\nReference file  : ",srefilename);
		if (fActor) {
			GetInput(scompfilename,"Enter name of .actors file to compare");
			fCheckCode = GetAnswer("Check for differences in codes? (Y/N)",'Y','N');
			fout << "Comparison file : " << scompfilename << endl;
			MergeFiles.mergeActors(scompfilename);
		}
		else {
			GetInput(scompfilename,"Enter name of .verbs file to compare");
			fCheckCode = GetAnswer("Check for differences in codes? (Y/N)",'Y','N');
			fout << "Comparison file : " << scompfilename << endl;
			MergeFiles.mergeVerbs(scompfilename);
		}
		fMore = GetAnswer("Check another file? (Y/N)",'Y','N');
	}
	WriteLine("Finished.  Your output is in the file ",sfout);
	fout.close();
} // doMerge

void MergeFilesClass:: mergeActors(char *sfilename)
// merge routine for .actors file
{
toktype index;
instring sc;

	ReadFiles.openFile(sfilename);
	while (ReadFiles.readLine()) {
		if (ReadFiles.fPhraseOK) {
			if (index = compareRoots(ReadFiles.sPhrase, Actor)) {   // deliberate use of assignment
				Roots.rootArray[index].used = 1;
				if (fCheckCode) {
					CodeStore.getActorString(sc,Roots.rootArray[index].ppat->icode);
					TrimEndBlanks(ReadFiles.sCode);
					if (strcmp(ReadFiles.sCode,sc)) {
						WriteLine("New code : ",ReadFiles.sPhrase);				
						fout << "\nNew code : " << ReadFiles.sPhrase <<endl;				
						fout << ">>> Old  : " << sc << endl;				
						fout << ">>> New  : " << ReadFiles.sCode <<endl;
					}
				}				
			}
			else {
				fout << "\nNew actor: "<< ReadFiles.sPhrase << " " << ReadFiles.sCode << " ;" << ReadFiles.sComment <<endl;			
				WRITESTRING("New actor: ");
				WRITESTRING(ReadFiles.sPhrase);
				WRITESTRING(" ");
				WRITESTRING(ReadFiles.sCode);
				WriteLine(" ;",ReadFiles.sComment);			
			}
		}
	}
	ReadFiles.fin.close();
	writeActorsMissing();
} // mergeActors

void MergeFilesClass:: mergeVerbs(char *sfilename)
// merge routine for .verbs file
{
toktype index, idxroot;
instring sc;
instring sphr;
patStruct *pPat;
instring star;
char *ps;
bool fshowRoot;				// true if root hasn't been written yet 
bool found;

	ReadFiles.openFile(sfilename);
	while (ReadFiles.readLine()) {
		if (ReadFiles.fPhraseOK) {
			if (ReadFiles.sPhrase[0] != '-') {			// deal with the root
				if (index = compareRoots(ReadFiles.sPhrase, Verb)) {   // deliberate use of assignment
					idxroot = index;
					Roots.rootArray[index].used = 1;
					fshowRoot = true;
					if (fCheckCode) {
						CodeStore.getEventString(sc,Roots.rootArray[index].ppat->icode);
						TrimEndBlanks(ReadFiles.sCode);
						if (strcmp(ReadFiles.sCode,sc)) {
							if (fshowRoot) {
								WriteLine("Verb      : ",ReadFiles.sPhrase);	
								fout << "\nVerb      : " << ReadFiles.sPhrase <<endl;	
								fshowRoot = false;
							}			
							fout << ">>Old code: " << sc <<endl;				
							fout << ">>New code: " << ReadFiles.sCode <<endl;
						}	
					}			
				}
				else {
					WRITESTRING("New verb  : ");
					WRITESTRING(ReadFiles.sPhrase);
					WRITESTRING(" ");
					WRITESTRING(ReadFiles.sCode);
					WriteLine(" ;",ReadFiles.sComment);			
					fout << "\nNew verb  : "<< ReadFiles.sPhrase << " " << ReadFiles.sCode << " ;" << ReadFiles.sComment <<endl;			
					idxroot = 0;
				}
			}  // if (sPhrase[0])
			else {  																// deal with phrase
				if (idxroot) {  // existing root; check the patterns list
					strcpy(star,&ReadFiles.sPhrase[2]);  // skip the "-"
					ps = star;
					while (*ps) ++ps;				// add blank unless string ends in '_'
					if ('_' != *(ps-1)) {
						*ps++ = ' ';
						*ps = '\0';
					}
					pPat = Roots.rootArray[index].ppat->pnext;	// go through patterns list
					found = false;
					while (pPat) {
						Phrases.get(sc,pPat->phrase); // get the phrase
						if (!strcmp(star,sc)) {   // 				&[2] skips over "- " in the comparison
							found = true;
							pPat->used = 1;
							if (fCheckCode) {
								CodeStore.getEventString(sc,pPat->icode);  // compare the phrase codes
								TrimEndBlanks(ReadFiles.sCode);
								if (strcmp(ReadFiles.sCode,sc)) {
									fout << "\nPhrase  : " << ReadFiles.sPhrase <<endl;				
									fout << ">>Old code: " << sc <<endl;				
									fout << ">>New code: " << ReadFiles.sCode <<endl;
								}
							}				
							break;
						}
						pPat = pPat->pnext;		// go to next pattern
					} // while pPat
					if (!found) {
						if (fshowRoot) {
							Phrases.get(sphr,Roots.rootArray[index].phrase);  // get the phrase
							WriteLine("Verb      : ",sphr);	
							fout << "\nVerb      : " << sphr <<endl;	
							fshowRoot = false;
						}			
						WRITESTRING("New phrase: ");
						WRITESTRING(ReadFiles.sPhrase);
						WRITESTRING(" ");
						WRITESTRING(ReadFiles.sCode);
						WriteLine(" ;",ReadFiles.sComment);			
						fout << "New phrase: "<< ReadFiles.sPhrase << " " << ReadFiles.sCode << " ;" << ReadFiles.sComment <<endl;			
					}
			}		// if idxroot
				else {  // phrase from new root, so just write it
					fout << "            : "<< ReadFiles.sPhrase << " " << ReadFiles.sCode << " ;" << ReadFiles.sComment <<endl;			
				}
			}  // else sPhrase[0] 
		} // if fPhraseOK
	}  // while readLine
	ReadFiles.fin.close();
	writeVerbsMissing();
} // mergeVerbs

toktype MergeFilesClass:: compareRoots(char *s, wordtype wtype)
// compare s to the existing dictionary
{
litStruct *plit;
toktype itok;
instring sphr;
instring star;
char *ps = star;
RootsClass::rootStruct root;

	strcpy(star,s);
	while (*ps) ++ps;				// add blank unless string ends in '_'
	if ('_' != *(ps-1)) {
		*ps++ = ' ';
		*ps = '\0';
	}
	Phrases.make(listAr,star);		// ### need a bounds check on listAr; maybe in make()?
//	cout << "MFC:cR: Target= '"<< s << "'");
	if (plit = Literals.litIndex[Literals.iHash(s)]) {  // deliberate use of assignment
		while (plit) {						// go through litIndex list
			if (listAr[0] == Literals.getIndex(plit)) {
			itok = plit->istart;		// go through istart list
			while (itok) {
				root =  Roots.rootArray[Literals.litBuf[itok]];
				if (wtype == root.wtype) {
					Phrases.get(sphr,root.phrase);  // get the phrase
//					cout << "MFC:cR: " << itok << "  '" << sphr << "'\n";
					if (!strcmp(sphr,star)) return Literals.litBuf[itok];   // found it
//					else cout << "Missed...\n";
				}			
				itok = Literals.litBuf[itok+1];		// continue through the istart list
			}	// while itok
		} 	// if listAr
		plit = plit->pnext;	// continue through litIndex list
		}		// while plit
	} 		// if litIndex
	return 0;					// not found

} // compareRoots

void MergeFilesClass:: writeActorsMissing(void)
// writes the .actors with use statistics to actorFile.use
{
	int ka;
	litStruct *plit;
	toktype itok;
	instring sphr,scode;
	RootsClass::rootStruct root;
	
	fout << "\n\nActors found in  '" << srefilename << "' and missing in '" << scompfilename << "'\n\n";
	WRITESTRING("\n\nActors found in  ");
	WRITESTRING(srefilename); 
	WriteLine(" and missing in ",scompfilename);
	for (ka = 0; ka<MAX_LIT_HASH; ka++) {   // go through litArray in alphabetical order
		if (plit = Literals.litIndex[ka]) {  // deliberate use of assignment
			while (plit) {						// go through litIndex list
				itok = plit->istart;		// go through istart list
				while (itok) {
					root =  Roots.rootArray[Literals.litBuf[itok]];
					if ((Actor == root.wtype) && !(root.used)) {
						Phrases.get(sphr,root.phrase);  // get the phrase
						CodeStore.getActorString(scode,root.ppat->icode);  // get code
						fout << sphr << "  " << scode << endl;
						WRITESTRING(sphr);
						WriteLine("  ",scode);
					}
					itok = Literals.litBuf[itok+1];		// continue through the istart list
				}	// while itok
			plit = plit->pnext;	// continue through litIndex list
			}	// while plit
		} // if litIndex
	} // for ka
} // writeActorsMissing

void MergeFilesClass:: writeVerbsMissing(void)
// writes the .verbs dictionary useage to sfilename.use
{
	int ka;
	litStruct *plit;
	toktype itok;
	instring sphr,scode,spat;
	patStruct *pPat;
	RootsClass::rootStruct root;
	bool fmiss;

	fout << "\n\nVerbs found in  '" << srefilename << "' and missing in '" << scompfilename << "'\n\n";
	WRITESTRING("\n\nVerbs found in  ");
	WRITESTRING(srefilename);
	WriteLine(" and missing in ",scompfilename);
	for (ka = 0; ka<MAX_LIT_HASH; ka++) {   // go through litArray in alphabetical order
		if (plit = Literals.litIndex[ka]) {  // deliberate use of assignment
			while (plit) {						// go through litIndex list
				itok = plit->istart;		// go through istart list
				while (itok) {
					root =  Roots.rootArray[Literals.litBuf[itok]];
					if (Verb == root.wtype) {
						fmiss = false;
						if (root.used) { // see if a pattern is unused						
							pPat = root.ppat->pnext;	// go through patterns list
							while (pPat) {
								if (!(pPat->used)) {
									fmiss = true;
									break;
								}
								pPat = pPat->pnext;
							} // while pPat
						}
						else fmiss = true;

						if (fmiss) {
							Phrases.get(sphr,root.phrase);  // get the phrase
							if (root.used) { 
								fout << sphr << " phrases missing\n";
								WriteLine(sphr," phrases missing");
							}
							else {
								CodeStore.getEventString(scode,root.ppat->icode);
								fout << sphr << "  " << scode << endl;
								WRITESTRING(sphr);
								WriteLine("  ",scode);
							}
							pPat = root.ppat->pnext;	// go through patterns list
							while (pPat) {
								if (!(pPat->used)) {
									Phrases.get(sphr,pPat->phrase); // get the phrase
									strcpy(spat,"- ");							// add the leading "- "
									strcat(spat,sphr);
									CodeStore.getEventString(scode,pPat->icode);
									fout << sphr << "  " << scode << endl;
									WRITESTRING(sphr);
									WriteLine("  ",scode);
								}
								pPat = pPat->pnext;
							} // while pPat
						fout << endl;					
						}
					}  // if Verb
					itok = Literals.litBuf[itok+1];		// continue through the istart list
				}	// while itok
			plit = plit->pnext;	// continue through litIndex list
			}	// while plit
		} // if litIndex
	} // for ka	
} // writeVerbsMissing

//___________________________________________________________________________________
// 																																		debugging code

#if FALSE // merge dictionaries test
	strcpy(sfout,"test.merge");
	fout.open(sfout,ios::out); 
	ReadFiles.readActors("Kv3.actors");
	MergeFiles.mergeActors("Kv3.actors.compare");
//	fout.open("test.verbs",ios::out); 
//	ReadFiles.readVerbs("Kv3.verbs");
//	MergeFiles.mergeVerbs("Kv3.verbs.compare");
	MergeFiles.doMerge();
#endif
